import jax
import jax.numpy as jnp
from flax import linen as nn
from typing import Sequence, Tuple

# --- Custom InstanceNorm1d Implementation ---
class InstanceNorm1d(nn.Module):
    """
    Flax implementation of Instance Normalization for 1D data (NLC format).
    Normalizes across the 'L' dimension.
    """
    features: int
    epsilon: float = 1e-5
    use_scale: bool = True
    use_bias: bool = True

    @nn.compact
    def __call__(self, x):
        """
        Args:
            x: Input tensor of shape (batch, length, features)

        Returns:
            Normalized tensor.
        """
        if x.shape[-1] != self.features:
             raise ValueError(f"Input features {x.shape[-1]} does not match InstanceNorm1d features {self.features}")

        # Calculate mean and variance across the length dimension (axis=1)
        # Keep dims for broadcasting
        mean = jnp.mean(x, axis=1, keepdims=True)
        # Variance calculation using mean needs care for numerical stability if needed,
        # but jnp.var should handle it.
        var = jnp.var(x, axis=1, keepdims=True)

        # Normalize
        normalized = (x - mean) / jnp.sqrt(var + self.epsilon)

        # Apply learnable scale and bias if enabled
        if self.use_scale:
            # Parameter shape: (features,) to broadcast across N and L
            scale = self.param('scale', nn.initializers.ones, (self.features,))
            normalized *= scale
        if self.use_bias:
            # Parameter shape: (features,)
            bias = self.param('bias', nn.initializers.zeros, (self.features,))
            normalized += bias

        return normalized

# --- Converted Modules ---

class ConvBlock(nn.Module):
    """Equivalent of the PyTorch conv_block function."""
    out_channels: int
    kernel_size: int = 3
    dilation: int = 1

    @nn.compact
    def __call__(self, x):
        """
        Args:
            x: Input tensor (N, L, C_in)
        Returns:
            Output tensor (N, L, C_out)
        """
        # Flax Conv expects kernel_size and dilation as sequences (tuples)
        ks = (self.kernel_size,)
        di = (self.dilation,)

        # Padding='SAME' attempts to preserve the length dimension for stride=1
        x = nn.Conv(
            features=self.out_channels,
            kernel_size=ks,
            kernel_dilation=di,
            padding='SAME'
        )(x)
        x = InstanceNorm1d(features=self.out_channels)(x) # Use custom InstanceNorm
        x = nn.PReLU()(x) # PReLU learns 'alpha' parameter per channel
        return x

class AttentionBlock(nn.Module):
    """Simple Channel Attention Block in Flax."""
    channels: int

    @nn.compact
    def __call__(self, x):
        """
        Args:
            x: Input tensor (N, L, C)
        Returns:
            Attention-weighted output tensor (N, L, C)
        """
        # Flax Conv expects kernel_size as a sequence (tuple)
        ks1 = (1,)
        attention_weights = nn.Conv(
            features=self.channels // 4, kernel_size=ks1, padding='SAME'
        )(x)
        # NOTE: PyTorch used inplace=True, JAX/Flax don't modify inplace
        attention_weights = nn.relu(attention_weights)
        attention_weights = nn.Conv(
            features=self.channels, kernel_size=ks1, padding='SAME'
        )(attention_weights)
        attention_weights = nn.sigmoid(attention_weights)

        return x * attention_weights

class ResidualInResidualBlock(nn.Module):
    """ResidualInResidualBlock in Flax."""
    channels: int
    num_convs: int = 3

    @nn.compact
    def __call__(self, x):
        """
        Args:
            x: Input tensor (N, L, C)
        Returns:
            Output tensor (N, L, C)
        """
        residual = x
        y = x
        # Sequentially apply ConvBlocks
        for _ in range(self.num_convs):
            y = ConvBlock(
                out_channels=self.channels,
                kernel_size=3, # Assuming kernel_size 3 as in original conv_block default
                dilation=1    # Assuming dilation 1 as in original conv_block default
            )(y)

        y = AttentionBlock(channels=self.channels)(y)
        return y + residual

class SISUGenerator(nn.Module):
    """SISUGenerator model translated to Flax."""
    channels: int = 16
    num_rirb: int = 4
    alpha: float = 1.0 # Non-learnable parameter, passed during init

    @nn.compact
    def __call__(self, x):
        """
        Args:
            x: Input tensor (N, L, 1) - assumes single channel input
        Returns:
            Output tensor (N, L, 1)
        """
        if x.shape[-1] != 1:
            raise ValueError(f"Input should have 1 channel (NLC format), got shape {x.shape}")

        residual_input = x

        # Initial convolution block
        # Flax Conv expects kernel_size as sequence
        ks7 = (7,)
        ks3 = (3,)
        y = nn.Conv(features=self.channels, kernel_size=ks7, padding='SAME')(x)
        y = InstanceNorm1d(features=self.channels)(y)
        y = nn.PReLU()(y)

        # Residual-in-Residual Blocks
        rirb_out = y
        for _ in range(self.num_rirb):
            rirb_out = ResidualInResidualBlock(channels=self.channels)(rirb_out)

        # Final layer
        learned_residual = nn.Conv(
            features=1, kernel_size=ks3, padding='SAME'
        )(rirb_out)

        # Combine with input residual
        output = residual_input + self.alpha * learned_residual
        return output
