import jax
import jax.numpy as jnp
import optax
import tqdm
import pickle # Using pickle for simplicity to save JAX states
import os
import argparse
# You might need a JAX-compatible library for audio loading/saving or convert to numpy
import scipy.io.wavfile as wavfile # Example for saving audio

import torch
from torch.utils.data import DataLoader

import file_utils as Data
from data import AudioDatasetNumPy
from generator import SISUGenerator
from discriminator import SISUDiscriminator

# Init script argument parser
parser = argparse.ArgumentParser(description="Training script")
parser.add_argument("--generator", type=str, default=None,
                    help="Path to the generator model file")
parser.add_argument("--discriminator", type=str, default=None,
                    help="Path to the discriminator model file")
parser.add_argument("--epoch", type=int, default=0, help="Starting epoch for model versioning")
parser.add_argument("--debug", action="store_true",  help="Print debug logs")
parser.add_argument("--continue_training", action="store_true", help="Continue training using temp_generator and temp_discriminator models")

args = parser.parse_args()

# Parameters
sample_rate = 44100
n_fft = 2048
hop_length = 256
win_length = n_fft
n_mels = 128
n_mfcc = 20 # If using MFCC

debug = args.debug

# Initialize JAX random key
key = jax.random.PRNGKey(0)

# Initialize dataset and dataloader
dataset_dir = './dataset/good'
dataset = AudioDatasetNumPy(dataset_dir) # Use your JAX dataset
train_data_loader = DataLoader(dataset, batch_size=4, shuffle=True) # Use your JAX DataLoader

models_dir = "models"
os.makedirs(models_dir, exist_ok=True)
audio_output_dir = "output"
os.makedirs(audio_output_dir, exist_ok=True)


# ========= MODELS =========

try:
    # Fetch the first batch
    first_batch = next(iter(train_data_loader))
    # The batch is a tuple: ((high_quality_audio_np, high_quality_sample_rate), (low_quality_audio_np, low_quality_sample_rate))
    # We need the high-quality audio NumPy array batch for initialization
    sample_input_np = first_batch[0][0] # Get the high-quality audio NumPy array batch
    # Convert the NumPy array batch to a JAX array
    sample_input_array = jnp.array(sample_input_np)

    # === FIX ===
    # Transpose the array from (batch, channels, length) to (batch, length, channels)
    # The original shape from DataLoader is likely (batch, channels, length) like (4, 1, 44100)
    # The generator expects NLC format (batch, length, channels) i.e., (4, 44100, 1)
    if sample_input_array.ndim == 3 and sample_input_array.shape[1] == 1:
         sample_input_array = jnp.transpose(sample_input_array, (0, 2, 1)) # Swap axes 1 and 2

    print(sample_input_array.shape) # Should now print (4, 44100, 1)
    # === END FIX ===

except StopIteration:
    print("Error: Data loader is empty. Cannot initialize models.")
    exit() # Exit if no data is available


key, init_key_g, init_key_d = jax.random.split(key, 3)
generator_model = SISUGenerator()
discriminator_model = SISUDiscriminator()

# Initialize parameters
generator_params = generator_model.init(init_key_g, sample_input_array)['params']
discriminator_params = discriminator_model.init(init_key_d, sample_input_array)['params']

# Define apply functions
generator_apply_fn = generator_model.apply
discriminator_apply_fn = discriminator_model.apply


# Loss functions (JAX equivalents)
# Optax provides common loss functions. BCEWithLogitsLoss is equivalent to
# sigmoid_binary_cross_entropy in Optax combined with a sigmoid activation
# in the model output or handling logits directly. Assuming your discriminator
# outputs logits, optax.sigmoid_binary_cross_entropy is appropriate.
criterion_d = optax.sigmoid_binary_cross_entropy
criterion_l1 = optax.sigmoid_binary_cross_entropy # For Mel, STFT, MFCC losses

# Optimizers (using Optax)
optimizer_g = optax.adam(learning_rate=0.0001, b1=0.5, b2=0.999)
optimizer_d = optax.adam(learning_rate=0.0001, b1=0.5, b2=0.999)

# Initialize optimizer states
generator_opt_state = optimizer_g.init(generator_params)
discriminator_opt_state = optimizer_d.init(discriminator_params)

# Schedulers - Optax has learning rate schedules. ReduceLROnPlateau
# is stateful and usually handled outside the jitted training step,
# or you can implement a custom learning rate schedule in Optax that
# takes a metric. For simplicity here, we won't directly replicate the
# PyTorch ReduceLROnPlateau but you could add logic based on losses
# in the main loop to adjust the learning rate if needed.


# Load saved state if continuing training
start_epoch = args.epoch
if args.continue_training:
    try:
        with open(f"{models_dir}/temp_generator.pkl", 'rb') as f:
            loaded_state = pickle.load(f)
            generator_params = loaded_state['params']
            generator_opt_state = loaded_state['opt_state']
        with open(f"{models_dir}/temp_discriminator.pkl", 'rb') as f:
            loaded_state = pickle.load(f)
            discriminator_params = loaded_state['params']
            discriminator_opt_state = loaded_state['opt_state']
        epoch_data = Data.read_data(f"{models_dir}/epoch_data.json")
        start_epoch = epoch_data.get("epoch", 0) + 1
        print(f"Continuing training from epoch {start_epoch}")
    except FileNotFoundError:
        print("Continue training requested but temp models not found. Starting from scratch.")
    except Exception as e:
        print(f"Error loading temp models: {e}. Starting from scratch.")

if args.generator is not None:
    try:
        with open(args.generator, 'rb') as f:
             loaded_state = pickle.load(f)
             generator_params = loaded_state['params']
        print(f"Loaded generator from {args.generator}")
    except FileNotFoundError:
        print(f"Generator model not found at {args.generator}")

if args.discriminator is not None:
    try:
        with open(args.discriminator, 'rb') as f:
            loaded_state = pickle.load(f)
            discriminator_params = loaded_state['params']
        print(f"Loaded discriminator from {args.discriminator}")
    except FileNotFoundError:
        print(f"Discriminator model not found at {args.discriminator}")


# Initialize JAX audio transforms
# mel_transform_fn = MelSpectrogramJAX(
#     sample_rate=sample_rate, n_fft=n_fft, hop_length=hop_length,
#     win_length=win_length, n_mels=n_mels, power=1.0 # Magnitude Mel
# )

# stft_transform_fn = SpectrogramJAX(
#     n_fft=n_fft, win_length=win_length, hop_length=hop_length
# )

# mfcc_transform_fn = MFCCJAX(
#     sample_rate,
#     n_mfcc,
#     melkwargs = {'n_fft': n_fft, 'hop_length': hop_length}
# )


# ========= JAX TRAINING STEPS =========

@jax.jit
def discriminator_train_step(
    discriminator_params,
    generator_params,
    discriminator_opt_state,
    high_quality_audio, # JAX array (batch, length, channels)
    low_quality_audio, # JAX array (batch, length, channels)
    real_labels, # JAX array
    fake_labels, # JAX array
    discriminator_apply_fn,
    generator_apply_fn,
    discriminator_optimizer,
    criterion_d,
    key # JAX random key
):
    # Split key for potential randomness in model application
    key, disc_key, gen_key = jax.random.split(key, 3)

    def loss_fn(d_params):
        # Generate fake audio
        # Note: Generator is not being trained in this step, so its parameters are static
        # Ensure low_quality_audio is in the expected NLC format (batch, length, channels)
        if low_quality_audio.ndim == 2: # Assuming (batch, length), add channel dim
             low_quality_audio = jnp.expand_dims(low_quality_audio, axis=-1)
        elif low_quality_audio.ndim == 1: # Assuming (length), add batch and channel dims
             low_quality_audio = jnp.expand_dims(jnp.expand_dims(low_quality_audio, axis=0), axis=-1)


        enhanced_audio, _ = generator_apply_fn({'params': generator_params}, gen_key, low_quality_audio)

        # Pass data through the discriminator
        # Ensure enhanced_audio has a leading dimension if not already present (e.g., batch size)
        if enhanced_audio.ndim == 2: # Assuming (length, channel), add batch dim
            enhanced_audio = jnp.expand_dims(enhanced_audio, axis=0)
        elif enhanced_audio.ndim == 1: # Assuming (length), add batch and channel dims
             enhanced_audio = jnp.expand_dims(jnp.expand_dims(enhanced_audio, axis=0), axis=-1)


        # Ensure high_quality_audio is in the expected NLC format (batch, length, channels)
        if high_quality_audio.ndim == 2: # Assuming (batch, length), add channel dim
             high_quality_audio = jnp.expand_dims(high_quality_audio, axis=-1)
        elif high_quality_audio.ndim == 1: # Assuming (length), add batch and channel dims
             high_quality_audio = jnp.expand_dims(jnp.expand_dims(high_quality_audio, axis=0), axis=-1)


        real_output = discriminator_apply_fn({'params': d_params}, disc_key, high_quality_audio)
        fake_output = discriminator_apply_fn({'params': d_params}, disc_key, enhanced_audio)

        # Calculate loss (criterion_d is assumed to be Optax's sigmoid_binary_cross_entropy or similar)
        # Ensure the shapes match the labels (batch_size, 1)
        real_output = real_output.reshape(-1, 1)
        fake_output = fake_output.reshape(-1, 1)

        real_loss = jnp.mean(criterion_d(real_output, real_labels))
        fake_loss = jnp.mean(criterion_d(fake_output, fake_labels))
        total_loss = real_loss + fake_loss
        return total_loss, (real_loss, fake_loss)

    # Compute gradients
    # Use jax.value_and_grad to get both the loss value and the gradients
    (loss, (real_loss, fake_loss)), grads = jax.value_and_grad(loss_fn, has_aux=True)(discriminator_params)

    # Apply updates
    updates, new_discriminator_opt_state = discriminator_optimizer.update(grads, discriminator_opt_state, discriminator_params)
    new_discriminator_params = optax.apply_updates(discriminator_params, updates)

    return new_discriminator_params, new_discriminator_opt_state, loss, key


@jax.jit
def generator_train_step(
    generator_params,
    discriminator_params,
    generator_opt_state,
    low_quality_audio, # JAX array (batch, length, channels)
    high_quality_audio, # JAX array (batch, length, channels)
    real_labels, # JAX array
    generator_apply_fn,
    discriminator_apply_fn,
    generator_optimizer,
    criterion_d, # Adversarial loss
    criterion_l1, # Feature matching loss
    key # JAX random key
):
    # Split key for potential randomness
    key, gen_key, disc_key = jax.random.split(key, 3)

    def loss_fn(g_params):
        # Ensure low_quality_audio is in the expected NLC format (batch, length, channels)
        if low_quality_audio.ndim == 2: # Assuming (batch, length), add channel dim
             low_quality_audio = jnp.expand_dims(low_quality_audio, axis=-1)
        elif low_quality_audio.ndim == 1: # Assuming (length), add batch and channel dims
             low_quality_audio = jnp.expand_dims(jnp.expand_dims(low_quality_audio, axis=0), axis=-1)

        # Generate enhanced audio
        enhanced_audio, _ = generator_apply_fn({'params': g_params}, gen_key, low_quality_audio)

        # Ensure enhanced_audio has a leading dimension if not already present
        if enhanced_audio.ndim == 2: # Assuming (length, channel), add batch dim
            enhanced_audio = jnp.expand_dims(enhanced_audio, axis=0)
        elif enhanced_audio.ndim == 1: # Assuming (length), add batch and channel dims
             enhanced_audio = jnp.expand_dims(jnp.expand_dims(enhanced_audio, axis=0), axis=-1)


        # Calculate adversarial loss (generator wants discriminator to think fake is real)
        # Note: Discriminator is not being trained in this step, so its parameters are static
        fake_output = discriminator_apply_fn({'params': discriminator_params}, disc_key, enhanced_audio)
        # Ensure the shape matches the labels (batch_size, 1)
        fake_output = fake_output.reshape(-1, 1)
        adversarial_loss = jnp.mean(criterion_d(fake_output, real_labels)) # Generator wants fake_output to be close to real_labels (1s)

        # Feature matching losses (assuming you add these back later)
        # You would need to implement JAX versions of your audio transforms
        # mel_loss = criterion_l1(mel_transform_fn(enhanced_audio), mel_transform_fn(high_quality_audio))
        # stft_loss = criterion_l1(stft_transform_fn(enhanced_audio), stft_transform_fn(high_quality_audio))
        # mfcc_loss = criterion_l1(mfcc_transform_fn(enhanced_audio), mfcc_transform_fn(high_quality_audio))

        # combined_loss = adversarial_loss + mel_loss + stft_loss + mfcc_loss
        combined_loss = adversarial_loss # For now, only adversarial loss

        # Return combined_loss and any other metrics needed for logging/analysis
        # For now, just adversarial loss and enhanced_audio
        return combined_loss, (adversarial_loss, enhanced_audio) # Add other losses here when implemented

    # Compute gradients
    # Update: loss_fn now returns (loss, (aux1, aux2, ...))
    (loss, (adversarial_loss_val, enhanced_audio)), grads = jax.value_and_grad(loss_fn, has_aux=True)(generator_params)

    # Apply updates
    updates, new_generator_opt_state = generator_optimizer.update(grads, generator_opt_state, generator_params)
    new_generator_params = optax.apply_updates(generator_params, updates)

    # Return the loss components separately along with the enhanced audio and key
    return new_generator_params, new_generator_opt_state, loss, adversarial_loss_val, enhanced_audio, key


# ========= MAIN TRAINING LOOP =========

def start_training():
    global generator_params, discriminator_params, generator_opt_state, discriminator_opt_state, key
    generator_epochs = 5000

    for generator_epoch in range(generator_epochs):
        current_epoch = start_epoch + generator_epoch

        # These will hold the last processed audio examples from a batch for saving
        last_high_quality_audio = None
        last_low_quality_audio = None
        last_ai_enhanced_audio = None
        last_sample_rate = None


        # Use tqdm for progress bar
        for high_quality_clip, low_quality_clip in tqdm.tqdm(train_data_loader, desc=f"Training epoch {generator_epoch+1}/{generator_epochs}, Current epoch {current_epoch}"):

            # high_quality_clip and low_quality_clip are tuples: (audio_array, sample_rate_array)
            # Extract audio arrays and sample rates (assuming batch dimension is first)
            # The arrays are NumPy arrays at this point, likely in (batch, channels, length) format
            high_quality_audio_batch_np = high_quality_clip[0]
            low_quality_audio_batch_np = low_quality_clip[0]
            sample_rate_batch_np = high_quality_clip[1] # Assuming sample rates are the same for paired clips

            # Convert NumPy arrays to JAX arrays and transpose to NLC format (batch, length, channels)
            # Only transpose if the shape is (batch, channels, length)
            if high_quality_audio_batch_np.ndim == 3 and high_quality_audio_batch_np.shape[1] == 1:
                 high_quality_audio_batch = jnp.transpose(jnp.array(high_quality_audio_batch_np), (0, 2, 1))
            else:
                 high_quality_audio_batch = jnp.array(high_quality_audio_batch_np) # Assume already NLC or handle other cases

            if low_quality_audio_batch_np.ndim == 3 and low_quality_audio_batch_np.shape[1] == 1:
                 low_quality_audio_batch = jnp.transpose(jnp.array(low_quality_audio_batch_np), (0, 2, 1))
            else:
                 low_quality_audio_batch = jnp.array(low_quality_audio_batch_np) # Assume already NLC or handle other cases

            sample_rate_batch = jnp.array(sample_rate_batch_np)


            batch_size = high_quality_audio_batch.shape[0]
            # Create labels - JAX arrays
            real_labels = jnp.ones((batch_size, 1))
            fake_labels = jnp.zeros((batch_size, 1))

            # Split key for each batch
            key, batch_key = jax.random.split(key)

            # ========= DISCRIMINATOR =========
            # Call the jitted discriminator training step
            discriminator_params, discriminator_opt_state, d_loss, batch_key = discriminator_train_step(
                discriminator_params,
                generator_params,
                discriminator_opt_state,
                high_quality_audio_batch,
                low_quality_audio_batch,
                real_labels,
                fake_labels,
                discriminator_apply_fn,
                generator_apply_fn,
                optimizer_d,
                criterion_d,
                batch_key
            )

            # ========= GENERATOR =========
            # Call the jitted generator training step
            generator_params, generator_opt_state, combined_loss, adversarial_loss, enhanced_audio_batch, batch_key = generator_train_step(
                generator_params,
                discriminator_params,
                generator_opt_state,
                low_quality_audio_batch,
                high_quality_audio_batch,
                real_labels, # Generator tries to make fake data look real
                generator_apply_fn,
                discriminator_apply_fn,
                optimizer_g,
                criterion_d,
                criterion_l1,
                batch_key
            )

            # Print debug logs (requires waiting for JIT compilation on first step)
            if debug:
                # Use .block_until_ready() to ensure computation is finished before printing
                # In a real scenario, you might want to log metrics less frequently
                d_loss_val = d_loss.block_until_ready().item()
                combined_loss_val = combined_loss.block_until_ready().item()
                adversarial_loss_val = adversarial_loss.block_until_ready().item()
                # Assuming other losses are returned by generator_train_step and unpacked
                # mel_loss_val = mel_l1_tensor.block_until_ready().item() if mel_l1_tensor is not None else 0
                # stft_loss_val = log_stft_l1_tensor.block_until_ready().item() if log_stft_l1_tensor is not None else 0
                # mfcc_loss_val = mfcc_l_tensor.block_until_ready().item() if mfcc_l_tensor is not None else 0
                print(f"D_LOSS: {d_loss_val:.4f}, G_COMBINED_LOSS: {combined_loss_val:.4f}, G_ADVERSARIAL_LOSS: {adversarial_loss_val:.4f}")
                # Print other losses here when implemented and returned

            # Schedulers - Implement your learning rate scheduling logic here if needed
            # based on the losses (e.g., reducing learning rate if loss plateaus).
            # This logic would typically live outside the jitted step function.
            # For Optax, you might use a schedule within the optimizer definition
            # or update the learning rate of the optimizer manually.


            # ========= SAVE LATEST AUDIO (from the last batch processed) =========
            # Access the first sample of the batch for saving
            # Ensure enhanced_audio_batch has a batch dimension and is in NLC format
            if enhanced_audio_batch.ndim == 2: # Assuming (length, channel), add batch dim
                enhanced_audio_batch = jnp.expand_dims(enhanced_audio_batch, axis=0)
            elif enhanced_audio_batch.ndim == 1: # Assuming (length), add batch and channel dims
                 enhanced_audio_batch = jnp.expand_dims(jnp.expand_dims(enhanced_audio_batch, axis=0), axis=-1)


            last_high_quality_audio = high_quality_audio_batch[0]
            last_low_quality_audio = low_quality_audio_batch[0]
            last_ai_enhanced_audio = enhanced_audio_batch[0]
            last_sample_rate = sample_rate_batch[0].item() # Assuming sample rate is scalar per batch item


        # Save audio files periodically (outside the batch loop)
        if generator_epoch % 25 == 0 and last_high_quality_audio is not None:
            print(f"Saving audio for epoch {current_epoch}!")
            try:
                # Convert JAX arrays to NumPy arrays for saving
                # Transpose back to (length, channels) or (length) if needed by wavfile.write
                # Assuming the models output (length, 1) or (length) after removing batch dim
                low_quality_audio_np_save = jax.device_get(last_low_quality_audio)
                ai_enhanced_audio_np_save = jax.device_get(last_ai_enhanced_audio)
                high_quality_audio_np_save = jax.device_get(last_high_quality_audio)

                # Remove the channel dimension if it's 1 for saving with wavfile
                if low_quality_audio_np_save.shape[-1] == 1:
                    low_quality_audio_np_save = low_quality_audio_np_save.squeeze(axis=-1)
                if ai_enhanced_audio_np_save.shape[-1] == 1:
                    ai_enhanced_audio_np_save = ai_enhanced_audio_np_save.squeeze(axis=-1)
                if high_quality_audio_np_save.shape[-1] == 1:
                    high_quality_audio_np_save = high_quality_audio_np_save.squeeze(axis=-1)


                wavfile.write(f"{audio_output_dir}/epoch-{current_epoch}-audio-crap.wav", last_sample_rate, low_quality_audio_np_save.astype(jnp.int16)) # Assuming audio is int16
                wavfile.write(f"{audio_output_dir}/epoch-{current_epoch}-audio-ai.wav", last_sample_rate, ai_enhanced_audio_np_save.astype(jnp.int16)) # Assuming audio is int16
                wavfile.write(f"{audio_output_dir}/epoch-{current_epoch}-audio-orig.wav", last_sample_rate, high_quality_audio_np_save.astype(jnp.int16)) # Assuming audio is int16
            except Exception as e:
                print(f"Error saving audio files: {e}")


        # Save model states periodically (outside the batch loop)
        # Use pickle to save parameters and optimizer states
        try:
            with open(f"{models_dir}/temp_discriminator.pkl", 'wb') as f:
                pickle.dump({'params': jax.device_get(discriminator_params), 'opt_state': jax.device_get(discriminator_opt_state)}, f)
            with open(f"{models_dir}/temp_generator.pkl", 'wb') as f:
                pickle.dump({'params': jax.device_get(generator_params), 'opt_state': jax.device_get(generator_opt_state)}, f)
            Data.write_data(f"{models_dir}/epoch_data.json", {"epoch": current_epoch})
        except Exception as e:
             print(f"Error saving temp model states: {e}")


    # Save final model states after all epochs
    print("Training complete! Saving final models.")
    try:
        with open(f"{models_dir}/epoch-{start_epoch + generator_epochs - 1}-discriminator.pkl", 'wb') as f:
            pickle.dump({'params': jax.device_get(discriminator_params)}, f)
        with open(f"{models_dir}/epoch-{start_epoch + generator_epochs - 1}-generator.pkl", 'wb') as f:
            pickle.dump({'params': jax.device_get(generator_params)}, f)
    except Exception as e:
        print(f"Error saving final model states: {e}")


start_training()
